import django

if django.VERSION < (3, 2, 0):
    # Only define default_app_config when using a version earlier than 3.2
    # https://docs.djangoproject.com/en/4.2/releases/3.2/#what-s-new-in-django-3-2
    default_app_config = "judoscale.django.apps.JudoscaleDjangoConfig"



import time
from typing import List, Optional

from judoscale.core.logger import logger
from judoscale.core.metric import Metric


class MetricsStore:
    """
    A thread-safe store for metrics.
    """

    def __init__(self, max_flush_interval: int = 60):
        self.store: List[Metric] = []
        # The max_flush_interval exists to prevent collecting metrics in memory
        # if they are never being flushed (if the reporter has failed for some
        # reason).
        self.max_flush_interval: int = max_flush_interval
        self.last_flush_time: float = time.monotonic()

    def add(self, metric: Metric) -> None:
        """
        Add a metric to the store.

        If the max flush interval has been exceeded, the metric will not be
        added to the store.
        """
        if self.last_flush_time > time.monotonic() - self.max_flush_interval:
            self.store.append(metric)
        else:
            logger.debug("Max flush interval exceeded - Not adding metric to store.")

    def flush(self) -> List[Metric]:
        """
        Return all metrics in the store and clear the store.
        """
        self.last_flush_time = time.monotonic()
        result = []
        # This operation needs to be atomic since the main thread is appending
        # to the store
        while metric := self._pop():
            result.append(metric)

        return result

    def _pop(self) -> Optional[Metric]:
        try:
            return self.store.pop() if self.store else None
        except IndexError:
            return None



import time
from datetime import timezone
from typing import List, Optional

from redis import Redis
from rq import Queue
from rq.job import Job

from judoscale.core.config import Config
from judoscale.core.logger import logger
from judoscale.core.metric import Metric
from judoscale.core.metrics_collectors import JobMetricsCollector

DEFAULTS = {
    "ENABLED": True,
    "MAX_QUEUES": 20,
    "QUEUES": [],
    "TRACK_BUSY_JOBS": False,
}


class RQMetricsCollector(JobMetricsCollector):
    def __init__(self, config: Config, redis: Redis):
        super().__init__(config=config)

        self.config["RQ"] = {**DEFAULTS, **self.config.get("RQ", {})}
        self.redis: Redis = redis
        logger.debug(f"Redis is at {self.redis.connection_pool}")
        logger.debug(f"Found initial queues: {list(self.queues)}")

    @property
    def adapter_config(self):
        return self.config["RQ"]

    @property
    def _queues(self) -> List[str]:
        return [q.name for q in Queue.all(connection=self.redis)]

    def oldest_job(self, queue: Queue) -> Optional[Job]:
        """
        Get the oldest task from the queue.
        """
        try:
            if jobs := queue.get_jobs(offset=0, length=1):
                return jobs[0]
        except Exception as e:
            logger.warning(f"Unable to get a task from queue: {queue.name}", exc_info=e)
        return None

    def collect(self) -> List[Metric]:
        metrics = []
        if not self.should_collect:
            return metrics

        logger.debug(f"Collecting metrics for queues {list(self.queues)}")
        queues = [Queue(name=q, connection=self.redis) for q in self.queues]

        for queue in queues:
            if self.adapter_config["TRACK_BUSY_JOBS"]:
                metrics.append(
                    Metric.for_busy_queue(
                        queue_name=queue.name,
                        busy_jobs=queue.started_job_registry.count,
                    )
                )

            if job := self.oldest_job(queue):
                if job.enqueued_at is not None:
                    # RQ stores `enqueued_at` as a naive datetime object, which
                    # means it doesn't have any timezone information associated
                    # with it.
                    # But since the time is, in fact, in UTC, we can just replace
                    # the timezone with UTC and then convert it to a timestamp.
                    # https://github.com/rq/rq/blob/42ac7d4150951b5f5fba8be11573920c1e6f6e69/rq/queue.py#L1003
                    #
                    enqueued_at = job.enqueued_at.replace(
                        tzinfo=timezone.utc
                    ).timestamp()

                    metrics.append(
                        Metric.for_queue(
                            queue_name=queue.name, oldest_job_ts=enqueued_at
                        )
                    )
            else:
                metrics.append(
                    Metric.for_queue(queue_name=queue.name, oldest_job_ts=time.time())
                )

        return metrics
